<?php

namespace UnicaenCode\Service;

use UnicaenCode\Manager\FormManager;
use UnicaenCode\Manager\HydratorManager;
use UnicaenCode\Manager\ServiceManager;
use UnicaenCode\Service\Traits\ConfigServiceAwareTrait;
use UnicaenCode\Util;
use UnicaenCode\Manager\ViewHelperManager;
use Zend\ModuleManager\ModuleManager;
use Zend\Mvc\Service\ViewHelperManagerFactory;

/**
 *
 *
 * @author Laurent LÉCLUSE <laurent.lecluse at unicaen.fr>
 */
class IntrospectionService
{
    use ConfigServiceAwareTrait;



    /**
     * Retourne la liste des services accessibles depuis le ServiceManager et correspondant aux critères suivants :
     * Les chaînes retournées sont les noms de classe des services
     *
     * @param null|string $namespace Les classes retournées devront faire partie du namespace spécifié
     * @param null|string $rootClass Les classes retournées devront hériter de $rootClass si spécifié
     *
     * @return string[]
     */
    public function getServices(?string $namespace = null, ?string $rootClass = null)
    {
        $manager = new ServiceManager(Util::getContainer());
        $result  = $manager->getClasses();

        return $this->filterClasses($result, $namespace, $rootClass);
    }



    /**
     * Retourne la liste des aides de vues accessibles depuis le ServiceManager et correspondant aux critères suivants :
     * Les chaînes retournées sont les noms de classe des aides de vues
     *
     * @param null|string $namespace Les classes retournées devront faire partie du namespace spécifié
     * @param null|string $rootClass Les classes retournées devront hériter de $rootClass si spécifié
     *
     * @return string[]
     */
    public function getViewHelpers(?string $namespace = null, ?string $rootClass = null): array
    {
        $manager = new ViewHelperManager(Util::getContainer());
        $result  = $manager->getClasses();

        return $this->filterClasses($result, $namespace, $rootClass);
    }



    /**
     * Retourne la liste des formulaires accessibles depuis le ServiceManager et correspondant aux critères suivants :
     * Les chaînes retournées sont les noms de classe des services
     *
     * @param null|string $namespace Les classes retournées devront faire partie du namespace spécifié
     * @param null|string $rootClass Les classes retournées devront hériter de $rootClass si spécifié
     *
     * @return string[]
     */
    public function getForms(?string $namespace = null, ?string $rootClass = null): array
    {
        $manager = new FormManager(Util::getContainer());
        $result  = $manager->getClasses();

        return $this->filterClasses($result, $namespace, $rootClass);
    }



    /**
     * Retourne la liste des entités utilisées par doctrine dans l'application
     *
     * @param null|string $namespace Les classes retournées devront faire partie du namespace spécifié
     * @param null|string $rootClass Les classes retournées devront hériter de $rootClass si spécifié
     *
     * @return string[]
     */
    public function getDbEntities(?string $namespace = null, ?string $rootClass = null): array
    {
        $result = [];
        $metas  = Util::getEntityManager()->getMetadataFactory()->getAllMetadata();
        foreach ($metas as $meta) {
            $eclass          = $meta->getName();
            $result[$eclass] = $eclass;
        }
        asort($result);

        return $this->filterClasses($result, $namespace, $rootClass);
    }



    /**
     * @param array       $services
     * @param string|null $namespace
     * @param string|null $rootClass
     *
     * @return array
     */
    private function filterClasses(array $classes, ?string $namespace = null, ?string $rootClass = null): array
    {
        if (substr($namespace, -1) !== '\\') {
            $namespace .= '\\';
        }
        foreach ($classes as $name => $class) {
            if ($namespace != '\\'){
                if (0 !== strpos($class, $namespace)){
                    unset($classes[$name]);
                }
            }

            if ($rootClass){
                if (!is_subclass_of($class, $rootClass)){
                    unset($classes[$name]);
                }
            }
        }

        return $classes;
    }



    /**
     * Retourne la liste des modules de l'application avec des informations associées
     *
     * Retourne un tableau associatif de la forme :
     *
     * nom du module => [
     *      name => string : nom du module,
     *      path => string : chemin d'accès du module,
     *      in-vendor => boolean : détermine si le module est dans /vendor ou non
     *      instance => instance du module (si nécessaire)
     * ]
     *
     * @param boolean $includeVendor inclue les modules du répertoire vendor ou non
     *
     * @return array
     */
    public function getModules($includeVendor = true)
    {
        $moduleManager = Util::getContainer()->get('ModuleManager');
        /* @var $moduleManager ModuleManager */

        $appPath      = getcwd();
        $modulesNames = $moduleManager->getModules();
        $modules      = [];
        foreach ($modulesNames as $module) {
            $instance = $moduleManager->getModule($module);
            $rc       = new \ReflectionClass($instance);

            $path     = dirname($rc->getFileName());
            $inVendor = 0 === strpos(str_replace($appPath . '/', '', $path), 'vendor');

            if (!$inVendor || $includeVendor) {
                $modules[$module] = [
                    'name'      => $module,
                    'path'      => $path,
                    'in-vendor' => $inVendor,
                    //      'instance' => $instance
                ];
            }
        }

        return $modules;
    }
}